/*
 * Clock and power management. This exposes the combined functionality
 * provided by the X1600 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-x1600.h"
#include <math.h>
#include <stdio.h>



// Register locations.

enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR
  Divider_ddr         = 0x02c,  // DDRCDR
  Divider_mac         = 0x054,  // MACCDR
  Divider0_i2s0       = 0x060,  // I2S0CDR
  Divider1_i2s0       = 0x070,  // I2S0CDR1
  Divider0_i2s1       = 0x07c,  // I2S1CDR (from X2000 manual)
  Divider1_i2s1       = 0x080,  // I2S1CDR1 (from X2000 manual)
  Divider_lcd         = 0x064,  // LPCDR
  Divider_msc0        = 0x068,  // MSC0CDR
  Divider_msc1        = 0x0a4,  // MSC1CDR
  Divider_sfc         = 0x074,  // SFCCDR
  Divider_ssi         = 0x05c,  // SSICDR
  Divider_cim         = 0x078,  // CIMCDR
  Divider_pwm         = 0x06c,  // PWMCDR
  Divider_can0        = 0x0a0,  // CAN0CDR
  Divider_can1        = 0x0a8,  // CAN1CDR
  Divider_cdbus       = 0x0ac,  // CDBUSCDR
  Divider_macphy0     = 0x0e4,  // MPHY0C
  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_swi             = 0x0bc,  // CPM_SFTINT
  Ddr_gate            = 0x0d0,  // DRCG
  Cpm_scratch_prot    = 0x038,  // CPSPPR
  Cpm_scratch         = 0x034,  // CPSPR
  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1
  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_fraction_A      = 0x084,  // CPAPACR
  Pll_fraction_M      = 0x088,  // CPMPACR
  Pll_fraction_E      = 0x08c,  // CPEPACR
};

enum Clock_source_values : unsigned
{
  Source_mME_main       = 0,
  Source_mME_pll_M      = 1,
  Source_mME_pll_E      = 2,

  // Special value

  Source_mask           = 0x3,
};



// Register field definitions.

static Field Clock_source_main         (Clock_control, 3, 30),       // SEL_SRC (output to SCLK_A)
             Clock_source_cpu          (Clock_control, 3, 28),       // SEL_CPLL (output to CCLK)
             Clock_source_hclock0      (Clock_control, 3, 26),       // SEL_H0PLL (output to AHB0)
             Clock_source_hclock2      (Clock_control, 3, 24),       // SEL_H2PLL (output to AHB2)
             Clock_source_can0         (Divider_can0, 3, 30),        // CA0CS
             Clock_source_can1         (Divider_can1, 3, 30),        // CA1CS
             Clock_source_cdbus        (Divider_cdbus, 3, 30),       // CDCS
             Clock_source_cim          (Divider_cim, 3, 30),         // CIMPCS
             Clock_source_ddr          (Divider_ddr, 3, 30),         // DCS
             Clock_source_i2s          (Divider0_i2s0, 1, 31),       // I2PCS
             Clock_source_lcd          (Divider_lcd, 3, 30),         // LPCS
             Clock_source_mac          (Divider_mac, 3, 30),         // MACPCS
             Clock_source_msc0         (Divider_msc0, 3, 30),        // MPCS
             Clock_source_msc1         (Divider_msc1, 3, 30),        // MPCS
             Clock_source_pwm          (Divider_pwm, 3, 30),         // PWMPCS
             Clock_source_sfc          (Divider_sfc, 3, 30),         // SFCS
             Clock_source_ssi          (Divider_ssi, 3, 30),         // SPCS

             Clock_busy_cpu            (Clock_status, 1, 0),
             Clock_busy_ddr            (Divider_ddr, 1, 28),
             Clock_busy_mac            (Divider_mac, 1, 28),
             Clock_busy_lcd            (Divider_lcd, 1, 28),
             Clock_busy_msc0           (Divider_msc0, 1, 28),
             Clock_busy_msc1           (Divider_msc1, 1, 28),
             Clock_busy_sfc            (Divider_sfc, 1, 28),
             Clock_busy_ssi            (Divider_ssi, 1, 28),
             Clock_busy_cim            (Divider_cim, 1, 28),
             Clock_busy_pwm            (Divider_pwm, 1, 28),
             Clock_busy_can0           (Divider_can0, 1, 28),
             Clock_busy_can1           (Divider_can1, 1, 28),
             Clock_busy_cdbus          (Divider_cdbus, 1, 28),

             Clock_change_enable_cpu   (Clock_control, 1, 22),
             Clock_change_enable_ahb0  (Clock_control, 1, 21),
             Clock_change_enable_ahb2  (Clock_control, 1, 20),
             Clock_change_enable_ddr   (Divider_ddr, 1, 29),
             Clock_change_enable_mac   (Divider_mac, 1, 29),
             Clock_change_enable_i2s   (Divider0_i2s0, 1, 29),
             Clock_change_enable_lcd   (Divider_lcd, 1, 29),
             Clock_change_enable_msc0  (Divider_msc0, 1, 29),
             Clock_change_enable_msc1  (Divider_msc1, 1, 29),
             Clock_change_enable_sfc   (Divider_sfc, 1, 29),
             Clock_change_enable_ssi   (Divider_ssi, 1, 29),
             Clock_change_enable_cim   (Divider_cim, 1, 29),
             Clock_change_enable_pwm   (Divider_pwm, 1, 29),
             Clock_change_enable_can0  (Divider_can0, 1, 29),
             Clock_change_enable_can1  (Divider_can1, 1, 29),
             Clock_change_enable_cdbus (Divider_cdbus, 1, 29),

             Clock_divider_can0        (Divider_can0, 0xff, 0),      // CAN0CDR
             Clock_divider_can1        (Divider_can1, 0xff, 0),      // CAN1CDR
             Clock_divider_cdbus       (Divider_cdbus, 0xff, 0),     // CDBUSCDR
             Clock_divider_cim         (Divider_cim, 0xff, 0),       // CIMCDR
             Clock_divider_cpu         (Clock_control, 0x0f, 0),     // CDIV
             Clock_divider_ddr         (Divider_ddr, 0x0f, 0),       // DDRCDR
             Clock_divider_hclock0     (Clock_control, 0x0f, 8),     // H0DIV (fast AHB peripherals)
             Clock_divider_hclock2     (Clock_control, 0x0f, 12),    // H2DIV (fast AHB peripherals)
             Clock_divider_i2s0_m      (Divider0_i2s0, 0x1ff, 20),   // I2SDIV_M
             Clock_divider_i2s0_n      (Divider0_i2s0, 0xfffff, 0),  // I2SDIV_N
             Clock_divider_i2s0_d      (Divider1_i2s0, 0xfffff, 0),  // I2SDIV_D
             Clock_divider_i2s1_m      (Divider0_i2s1, 0x1ff, 20),   // I2SDIV_M
             Clock_divider_i2s1_n      (Divider0_i2s1, 0xfffff, 0),  // I2SDIV_N
             Clock_divider_i2s1_d      (Divider1_i2s1, 0xfffff, 0),  // I2SDIV_D
             Clock_divider_l2cache     (Clock_control, 0x0f, 4),     // L2CDIV
             Clock_divider_lcd         (Divider_lcd, 0xff, 0),       // LPCDR
             Clock_divider_mac         (Divider_mac, 0xff, 0),       // MACCDR
             Clock_divider_msc0        (Divider_msc0, 0xff, 0),      // MSC0CDR
             Clock_divider_msc1        (Divider_msc1, 0xff, 0),      // MSC1CDR
             Clock_divider_pclock      (Clock_control, 0x0f, 16),    // PDIV (slow APB peripherals)
             Clock_divider_pwm         (Divider_pwm, 0x0f, 0),       // PWMCDR
             Clock_divider_sfc         (Divider_sfc, 0xff, 0),       // SFCCDR
             Clock_divider_ssi         (Divider_ssi, 0xff, 0),       // SSICDR

             Clock_gate_main           (Clock_control, 1, 23),       // GATE_SCLKA
             Clock_gate_ddr            (Clock_gate0, 1, 31),         // DDR
             Clock_gate_ahb0           (Clock_gate0, 1, 29),         // AHB0
             Clock_gate_apb0           (Clock_gate0, 1, 28),         // APB0
             Clock_gate_rtc            (Clock_gate0, 1, 27),         // RTC
             Clock_gate_aes            (Clock_gate0, 1, 24),         // AES
             Clock_gate_lcd_pixel      (Clock_gate0, 1, 23),         // LCD
             Clock_gate_cim            (Clock_gate0, 1, 22),         // CIM
             Clock_gate_dma            (Clock_gate0, 1, 21),         // PDMA
             Clock_gate_ost            (Clock_gate0, 1, 20),         // OST
             Clock_gate_ssi0           (Clock_gate0, 1, 19),         // SSI0
             Clock_gate_timer          (Clock_gate0, 1, 18),         // TCU
             Clock_gate_dtrng          (Clock_gate0, 1, 17),         // DTRNG
             Clock_gate_uart2          (Clock_gate0, 1, 16),         // UART2
             Clock_gate_uart1          (Clock_gate0, 1, 15),         // UART1
             Clock_gate_uart0          (Clock_gate0, 1, 14),         // UART0
             Clock_gate_sadc           (Clock_gate0, 1, 13),         // SADC
             Clock_gate_audio          (Clock_gate0, 1, 11),         // AUDIO
             Clock_gate_ssi_slv        (Clock_gate0, 1, 10),         // SSI_SLV
             Clock_gate_i2c1           (Clock_gate0, 1, 8),          // I2C1
             Clock_gate_i2c0           (Clock_gate0, 1, 7),          // I2C0
             Clock_gate_msc1           (Clock_gate0, 1, 5),          // MSC1
             Clock_gate_msc0           (Clock_gate0, 1, 4),          // MSC0
             Clock_gate_otg            (Clock_gate0, 1, 3),          // OTG
             Clock_gate_sfc            (Clock_gate0, 1, 2),          // SFC
             Clock_gate_efuse          (Clock_gate0, 1, 1),          // EFUSE
             Clock_gate_nemc           (Clock_gate0, 1, 0),          // NEMC
             Clock_gate_arb            (Clock_gate1, 1, 30),         // ARB
             Clock_gate_mipi_csi       (Clock_gate1, 1, 28),         // MIPI_CSI
             Clock_gate_intc           (Clock_gate1, 1, 26),         // INTC
             Clock_gate_gmac0          (Clock_gate1, 1, 23),         // GMAC0
             Clock_gate_uart3          (Clock_gate1, 1, 16),         // UART3
             Clock_gate_i2s0_tx        (Clock_gate1, 1, 9),          // I2S0_dev_tclk
             Clock_gate_i2s0_rx        (Clock_gate1, 1, 8),          // I2S0_dev_rclk
             Clock_gate_hash           (Clock_gate1, 1, 6),          // HASH
             Clock_gate_pwm            (Clock_gate1, 1, 5),          // PWM
             Clock_gate_cdbus          (Clock_gate1, 1, 2),          // CDBUS
             Clock_gate_can1           (Clock_gate1, 1, 1),          // CAN1
             Clock_gate_can0           (Clock_gate1, 1, 0),          // CAN0

             Pll_enable_A              (Pll_control_A, 1, 0),        // APLLEN
             Pll_enable_E              (Pll_control_E, 1, 0),        // EPLLEN
             Pll_enable_M              (Pll_control_M, 1, 0),        // MPLLEN

             Pll_stable_A              (Pll_control_A, 1, 3),        // APLL_ON
             Pll_stable_E              (Pll_control_E, 1, 3),        // EPLL_ON
             Pll_stable_M              (Pll_control_M, 1, 3),        // MPLL_ON

             Pll_bypass_A              (Pll_control_A, 1, 30),       // APLL_BP
             Pll_bypass_E              (Pll_control_E, 1, 26),       // EPLL_BP
             Pll_bypass_M              (Pll_control_M, 1, 28),       // MPLL_BP

             Pll_multiplier_A          (Pll_control_A, 0x1fff, 20),  // APLLM
             Pll_multiplier_E          (Pll_control_E, 0x1fff, 20),  // EPLLM
             Pll_multiplier_M          (Pll_control_M, 0x1fff, 20),  // MPLLM

             Pll_input_division_A      (Pll_control_A, 0x3f, 14),    // APLLN
             Pll_input_division_E      (Pll_control_E, 0x3f, 14),    // EPLLN
             Pll_input_division_M      (Pll_control_M, 0x3f, 14),    // MPLLN

             Pll_output_division1_A    (Pll_control_A, 0x07, 11),    // APLLOD1
             Pll_output_division1_E    (Pll_control_E, 0x07, 11),    // EPLLOD1
             Pll_output_division1_M    (Pll_control_M, 0x07, 11),    // MPLLOD1

             Pll_output_division0_A    (Pll_control_A, 0x07, 8),     // APLLOD0
             Pll_output_division0_E    (Pll_control_E, 0x07, 8),     // EPLLOD0
             Pll_output_division0_M    (Pll_control_M, 0x07, 8);     // MPLLOD0



// Multiplexer instances.

#define Clocks(...) ((enum Clock_identifiers []) {__VA_ARGS__})

static Mux mux_external (Clock_external),
           mux_pclock   (Clock_pclock),
           mux_ahb2_apb (Clock_ahb2_apb),
           mux_core     (3, Clocks(Clock_none, Clock_main, Clock_pll_M)),
           mux_bus      (4, Clocks(Clock_main, Clock_pll_M, Clock_pll_E, Clock_external)),
           mux_dev      (3, Clocks(Clock_main, Clock_pll_M, Clock_pll_E)),
           mux_i2s      (2, Clocks(Clock_main, Clock_pll_E));



// Clock instances.

static Clock_passive clock_external;

static Clock_null clock_none;

static Clock clock_ahb2_apb(Source(mux_core, Clock_source_hclock2)),

             clock_can0(Source(mux_bus, Clock_source_can0),
                        Control(Clock_gate_can0, Clock_change_enable_can0, Clock_busy_can0),
                        Divider(Clock_divider_can0)),
 
             clock_can1(Source(mux_bus, Clock_source_can1),
                        Control(Clock_gate_can1, Clock_change_enable_can1, Clock_busy_can1),
                        Divider(Clock_divider_can1)),
 
             clock_cdbus(Source(mux_dev, Clock_source_cdbus),
                         Control(Clock_gate_cdbus, Clock_change_enable_cdbus, Clock_busy_cdbus),
                         Divider(Clock_divider_cdbus)),
 
             clock_cim(Source(mux_dev, Clock_source_cim),
                       Control(Clock_gate_cim, Clock_change_enable_cim, Clock_busy_cim),
                       Divider(Clock_divider_cim)),
 
             clock_cpu(Source(mux_core, Clock_source_cpu),
                       Control(Field::undefined, Clock_change_enable_cpu, Clock_busy_cpu),
                       Divider(Clock_divider_cpu)),
 
             clock_ddr(Source(mux_core, Clock_source_ddr),
                       Control(Clock_gate_ddr, Clock_change_enable_ddr, Clock_busy_ddr),
                       Divider(Clock_divider_ddr)),
 
             clock_dma(Source(mux_pclock), Control(Clock_gate_dma), Divider::undefined),
 
             clock_hclock0(Source(mux_core, Clock_source_hclock0),
                           Control(Clock_gate_ahb0, Clock_change_enable_ahb0),
                           Divider(Clock_divider_hclock0)),
 
             clock_hclock2(Source(mux_ahb2_apb),
                           Control(Clock_gate_apb0, Clock_change_enable_ahb2),
                           Divider(Clock_divider_hclock2)),
 
             clock_i2c(Source(mux_pclock), Control(Clock_gate_i2c0), Divider::undefined),
 
             clock_i2c0(Source(mux_pclock), Control(Clock_gate_i2c0), Divider::undefined),
 
             clock_i2c1(Source(mux_pclock), Control(Clock_gate_i2c1), Divider::undefined),
 
             clock_lcd_pixel(Source(mux_dev, Clock_source_lcd),
                             Control(Clock_gate_lcd_pixel, Clock_change_enable_lcd, Clock_busy_lcd),
                             Divider(Clock_divider_lcd)),
 
             clock_mac(Source(mux_dev, Clock_source_mac),
                       Control(Clock_gate_gmac0, Clock_change_enable_mac, Clock_busy_mac),
                       Divider(Clock_divider_mac)),
 
             clock_main(Source(mux_core, Clock_source_main),
                        Control(Clock_gate_main)),
 
             clock_msc(Source(mux_dev, Clock_source_msc0),
                       Control(Clock_gate_msc0, Clock_change_enable_msc0, Clock_busy_msc0),
                       Divider(Clock_divider_msc0)),
 
             clock_msc0(Source(mux_dev, Clock_source_msc0),
                        Control(Clock_gate_msc0, Clock_change_enable_msc0, Clock_busy_msc0),
                        Divider(Clock_divider_msc0)),
 
             clock_msc1(Source(mux_dev, Clock_source_msc1),
                        Control(Clock_gate_msc1, Clock_change_enable_msc1, Clock_busy_msc1),
                        Divider(Clock_divider_msc1)),
 
             clock_pclock(Source(mux_ahb2_apb),
                          Control(Clock_gate_apb0, Field::undefined, Field::undefined),
                          Divider(Clock_divider_pclock)),
 
             clock_pwm(Source(mux_dev, Clock_source_pwm),
                       Control(Clock_gate_pwm, Clock_change_enable_pwm, Clock_busy_pwm),
                       Divider(Clock_divider_pwm)),
 
             clock_pwm0(Source(mux_dev, Clock_source_pwm),
                        Control(Clock_gate_pwm, Clock_change_enable_pwm, Clock_busy_pwm),
                        Divider(Clock_divider_pwm)),
 
             clock_sfc(Source(mux_dev, Clock_source_sfc),
                       Control(Clock_gate_sfc, Clock_change_enable_sfc, Clock_busy_sfc),
                       Divider(Clock_divider_sfc)),
 
             clock_ssi(Source(mux_dev, Clock_source_ssi),
                       Control(Clock_gate_ssi0, Clock_change_enable_ssi, Clock_busy_ssi),
                       Divider(Clock_divider_ssi)),
 
             clock_timer(Source(mux_pclock), Control(Clock_gate_timer), Divider::undefined),
 
             clock_uart0(Source(mux_external), Control(Clock_gate_uart0), Divider::undefined),
 
             clock_uart1(Source(mux_external), Control(Clock_gate_uart1), Divider::undefined),
 
             clock_uart2(Source(mux_external), Control(Clock_gate_uart2), Divider::undefined),
 
             clock_uart3(Source(mux_external), Control(Clock_gate_uart3), Divider::undefined);
 
static Clock_divided_i2s clock_i2s0_rx(Source(mux_i2s, Clock_source_i2s),
                                       Control(Clock_gate_i2s0_rx, Clock_change_enable_i2s),
                                       Divider_i2s(Clock_divider_i2s0_m, Clock_divider_i2s0_n,
                                                   Clock_divider_i2s0_d)),

                         clock_i2s0_tx(Source(mux_i2s, Clock_source_i2s),
                                       Control(Clock_gate_i2s0_tx, Clock_change_enable_i2s),
                                       Divider_i2s(Clock_divider_i2s1_m, Clock_divider_i2s1_n,
                                                   Clock_divider_i2s1_d));

static Pll clock_pll_A(Source(mux_external),
                       Control_pll(Pll_enable_A, Pll_stable_A, Pll_bypass_A),
                       Divider_pll(Pll_multiplier_A, Pll_input_division_A,
                                   Pll_output_division0_A, Pll_output_division1_A)),

           clock_pll_E(Source(mux_external),
                       Control_pll(Pll_enable_E, Pll_stable_E, Pll_bypass_E),
                       Divider_pll(Pll_multiplier_E, Pll_input_division_E,
                                   Pll_output_division0_E, Pll_output_division1_E)),

           clock_pll_M(Source(mux_external),
                       Control_pll(Pll_enable_M, Pll_stable_M, Pll_bypass_M),
                       Divider_pll(Pll_multiplier_M, Pll_input_division_M,
                                   Pll_output_division0_M, Pll_output_division1_M));



// Clock register.

static Clock_base *clocks[Clock_identifier_count] = {
  &clock_ahb2_apb,
  &clock_none,        // Clock_aic_bitclk
  &clock_none,        // Clock_aic_pclk
  &clock_can0,
  &clock_can1,
  &clock_cdbus,
  &clock_cim,
  &clock_cpu,
  &clock_ddr,
  &clock_dma,
  &clock_none,        // Clock_emac
  &clock_external,
  &clock_hclock0,
  &clock_hclock2,
  &clock_none,        // Clock_hdmi
  &clock_i2c,
  &clock_i2c0,
  &clock_i2c1,
  &clock_none,        // Clock_i2s
  &clock_i2s0_rx,
  &clock_i2s0_tx,
  &clock_none,        // Clock_kbc
  &clock_none,        // Clock_lcd
  &clock_lcd_pixel,
  &clock_mac,
  &clock_main,
  &clock_msc,
  &clock_msc0,
  &clock_msc1,
  &clock_none,
  &clock_pclock,
  &clock_pll_A,
  &clock_pll_E,
  &clock_pll_M,
  &clock_pwm,
  &clock_pwm0,
  &clock_none,        // Clock_pwm1
  &clock_none,        // Clock_scc
  &clock_sfc,
  &clock_none,        // Clock_smb0
  &clock_none,        // Clock_smb1
  &clock_none,        // Clock_smb2
  &clock_none,        // Clock_smb3
  &clock_none,        // Clock_smb4
  &clock_ssi,
  &clock_timer,
  &clock_uart0,
  &clock_uart1,
  &clock_uart2,
  &clock_uart3,
  &clock_none,        // Clock_udc
  &clock_none,        // Clock_uhc
  &clock_none,        // Clock_uprt
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_x1600_chip::Cpm_x1600_chip(l4_addr_t addr, uint32_t exclk_freq)
: _cpm_regs(addr, clocks, exclk_freq)
{
  // add_cid("cpm");
  // add_cid("cpm-x1600");
  // register_property("exclk_freq", &exclk_freq);
}

const char *
Cpm_x1600_chip::clock_type(enum Clock_identifiers clock)
{
  return clocks[clock]->clock_type();
}

int
Cpm_x1600_chip::have_clock(enum Clock_identifiers clock)
{
  return clocks[clock]->have_clock(_cpm_regs);
}

void
Cpm_x1600_chip::start_clock(enum Clock_identifiers clock)
{
  clocks[clock]->start_clock(_cpm_regs);
}

void
Cpm_x1600_chip::stop_clock(enum Clock_identifiers clock)
{
  clocks[clock]->stop_clock(_cpm_regs);
}

uint32_t
Cpm_x1600_chip::get_divider(enum Clock_identifiers clock)
{
  Clock *clk = dynamic_cast<Clock *>(clocks[clock]);

  if (clk != NULL)
    return clk->get_divider(_cpm_regs);
  else
    return 1;
}

void
Cpm_x1600_chip::set_divider(enum Clock_identifiers clock, uint32_t division)
{
  Clock *clk = dynamic_cast<Clock *>(clocks[clock]);

  if (clk != NULL)
    clk->set_divider(_cpm_regs, division);
}

uint8_t
Cpm_x1600_chip::get_source(enum Clock_identifiers clock)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(clocks[clock]);

  if (clk != NULL)
    return clk->get_source(_cpm_regs);
  else
    return 0;
}

void
Cpm_x1600_chip::set_source(enum Clock_identifiers clock, uint8_t source)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(clocks[clock]);

  if (clk != NULL)
    clk->set_source(_cpm_regs, source);
}

uint32_t
Cpm_x1600_chip::get_source_frequency(enum Clock_identifiers clock)
{
  Clock_active *clk = dynamic_cast<Clock_active *>(clocks[clock]);

  if (clk != NULL)
    return clk->get_source_frequency(_cpm_regs);
  else
    return 0;
}

uint32_t
Cpm_x1600_chip::get_frequency(enum Clock_identifiers clock)
{
  return clocks[clock]->get_frequency(_cpm_regs);
}

void
Cpm_x1600_chip::set_frequency(enum Clock_identifiers clock, uint32_t frequency)
{
  switch (clock)
  {
    // The pixel frequency is based on the selected clock source (SCLK_A, MPLL or
    // EPLL).

    case Clock_lcd_pixel:
    {

      // Switch to the MPLL and attempt to set the divider.

      Clock *lcd = dynamic_cast<Clock *>(clocks[Clock_lcd_pixel]);
      Clock_base *pll = clocks[Clock_pll_M];

      if (lcd != NULL)
      {
        lcd->set_source(_cpm_regs, Source_mME_pll_M);
        pll->start_clock(_cpm_regs);
        lcd->set_divider(_cpm_regs, lcd->get_source_frequency(_cpm_regs) / frequency);
      }
      break;
    }

    default:
      break;
  }
}

void
Cpm_x1600_chip::set_pll_parameters(enum Clock_identifiers clock, uint16_t multiplier,
                                   uint8_t in_divider, uint8_t out_divider)
{
  Pll *pll = dynamic_cast<Pll *>(clocks[clock]);

  if (pll != NULL)
    pll->set_parameters(_cpm_regs, multiplier, in_divider, out_divider);
}



// C language interface functions.

void
*x1600_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 24MHz EXCLK frequency. */

  return (void *) new Cpm_x1600_chip(cpm_base, 24000000);
}

const char *
x1600_cpm_clock_type(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->clock_type(clock);
}

int
x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->have_clock(clock);
}

void
x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->start_clock(clock);
}

void
x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->stop_clock(clock);
}

uint32_t
x1600_cpm_get_divider(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_divider(clock);
}

void
x1600_cpm_set_divider(void *cpm, enum Clock_identifiers clock, uint32_t divider)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->set_divider(clock, divider);
}

uint8_t
x1600_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source(clock);
}

void
x1600_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{       
  static_cast<Cpm_x1600_chip *>(cpm)->set_source(clock, source);
}

uint32_t
x1600_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source_frequency(clock);
}

uint32_t
x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_frequency(clock);
}

void
x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_frequency(clock, frequency);
}

void
x1600_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_pll_parameters(Clock_pll_M, multiplier, in_divider, out_divider);
}
